/*
Copyright (c) 2016 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <stdint.h>
#include <stdarg.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/extensions/XShm.h>
#include "graymap.h"
#include "txtren.h"
#include "x11iface.h"
#include "config.h"
#include "menu.h"
#include "ptrbuf.h"

struct xmenu
{
  Window parent, win;
  int w,h;
  ximage_t *image;
  int itemh;
  int sel,click;
  int n_items;

  void (*func)(int ,void*);
  void *arg;
  void *parent_click_handler,*parent_release_handler;
};

static void collect_varargs(va_list ap, void ***Rptr, size_t *Rn);
static void render_items(menu_t *M, char **items);
static void menuE_pointermove(menu_t *M,XMotionEvent *event);
static void menuE_buttonpress(menu_t *M,XButtonEvent *event);
static void menuE_buttonrelease(menu_t *M,XButtonEvent *event);
static void menuE_expose(menu_t *M,XExposeEvent *event);
static void menuE_parentclick(menu_t *M,XButtonEvent *event);
static void menuE_parentrelease(menu_t *M,XButtonEvent *event);
static void close_menu(menu_t *M, int idx);
static void hilight_item(menu_t *M, int x, int y);
static int  menu_i2y(menu_t* M, int idx);
static int  menu_y2i(menu_t* M, int y);
static void draw_hilight(menu_t *M,int idx);

static int MO_pad_hor= 10,
           MO_pad_ver= 6,
           MO_pad_mid= 6, 
           MO_bgcolor= 0x445544,
           MO_fgcolor= 0xeeffee,
           MO_hilight= 0xffFFff,
           MO_font_size= 10;
static char * MO_font_name= ":default.ttf";

static u_font_t *M_font;
static int M_fonth;

void menu_init(cfg_menu_style_t *cfg)
{
#define L(X)  if (cfg->G_##X) MO_##X= cfg->X;
  L(pad_hor) L(pad_ver) L(pad_mid) L(bgcolor) L(fgcolor) L(hilight)
  L(font_size) L(font_name)
#undef L
  if (M_font) font_free(M_font);
  M_font= font_new(MO_font_name, MO_font_size);
  if (!M_font) M_font= font_new(":default.ttf", 10);
  M_fonth= M_font->height;
}

menu_t *menu_new(void (*result)(int ,void *), void *arg, ...)
{
  menu_t *M;
  va_list ap;
  char **items;
  size_t n_items;

  M= calloc(1,sizeof(*M));
  M->func= result;
  M->arg= arg;

  va_start(ap,arg);
  collect_varargs(ap,(void***) &items, &n_items);
  va_end(ap);

  M->n_items= n_items;
  M->itemh= M_fonth + MO_pad_mid;

  render_items(M, items);
  free(items);

  return M;
}

static void render_items(menu_t *M, char **items)
{
  int i,y;
  uint8_t **td;
  uint32_t *tw;
  uint32_t maxw;
  int n_items;

  n_items= M->n_items;

  td= malloc(n_items*sizeof(td[0]));
  tw= malloc(n_items*sizeof(tw[0]));
  maxw= 0;

  for(i=0;i<n_items;i++)
  {
    text_render(M_font,items[i],tw+i,td+i);
    if (tw[i]>maxw) maxw= tw[i];
  }

  M->h= n_items*M->itemh + 2*MO_pad_ver;
  M->w= maxw + 2*MO_pad_hor;

  M->image= create_image(M->w,M->h);
  fill_rect(M->image->data,M->w,M->h,0,0,M->w,M->h,MO_bgcolor);

  y= MO_pad_ver + MO_pad_mid/2;

  for(i=0;i<n_items;i++)
  {
     graymap_draw(M->image->data, M->w, M->h,
              MO_pad_hor, y, 
              td[i], tw[i], M_fonth,
              0,0,
              MO_fgcolor, MO_bgcolor, tw[i], M_fonth);
     free(td[i]);
     y+= M->itemh;
  }
  free(td);
  free(tw);
}


static void collect_varargs(va_list ap, void ***Rptr, size_t *Rn)
{
  ptrbuf_t *B;
  void *K;
  B= ptrbuf_new(50);
  while((K=va_arg(ap,void*))) ptrbuf_put(B,K);
  ptrbuf_free(B, Rptr, Rn); 
}

void menu_display(menu_t *M, Window parent, int x, int y)
{
  XSetWindowAttributes attribs;

  M->parent= parent;
  attribs.background_pixmap= None;
  M->win= XCreateWindow(x_display, parent,
                     x, y, M->w, M->h,
                     0, x_depth, InputOutput, x_visual, 
                     CWBackPixmap, &attribs);
  XSelectInput(x_display, M->win, 
       ExposureMask | ButtonPressMask | ButtonReleaseMask | PointerMotionMask );

  register_event_handler(M->win, MotionNotify, M, menuE_pointermove, NULL);
  register_event_handler(M->win, ButtonPress, M, menuE_buttonpress, NULL);
  register_event_handler(M->win, ButtonRelease, M, menuE_buttonrelease, NULL);
  register_event_handler(M->win, Expose, M, menuE_expose, NULL);
  register_event_handler(M->parent, ButtonPress, M, menuE_parentclick,
                                                 &M->parent_click_handler);
  register_event_handler(M->parent, ButtonRelease, M, menuE_parentrelease,
                                                 &M->parent_release_handler);

  XMapWindow(x_display,M->win);
  M->sel= -1;
  M->click= -1;

  int px,py, DUMi;
  unsigned int DUMu;
  Window DUMw;

  XQueryPointer(x_display, M->win, &DUMw, &DUMw, &DUMi, &DUMi, &px, &py, &DUMu);
  hilight_item(M, px, py);
}

static void menuE_pointermove(menu_t *M,XMotionEvent *event)
{
  hilight_item(M,event->x,event->y);
}

static void menuE_buttonpress(menu_t *M,XButtonEvent *event)
{
  if (event->button!=1) return ;
  M->click= menu_y2i(M,event->y);
}

static void menuE_buttonrelease(menu_t *M,XButtonEvent *event)
{
  if (event->button!=1) return ;
  close_menu(M, M->click);
}

static void menuE_expose(menu_t *M,XExposeEvent *event)
{
  put_image(M->win, event->x,event->y,
            M->image, event->x, event->y,
            event->width, event->height);
}

static void menuE_parentclick(menu_t *M,XButtonEvent *event)
{
  // we actually do nothing here. the actual destruction of
  // the menu is done in the release handler. this is just
  // here to eat up the click event as well as the release
  // event for the button.
}
static void menuE_parentrelease(menu_t *M,XButtonEvent *event)
{
  close_menu(M, -1);
}

static void close_menu(menu_t *M, int idx)
{
  if (M->sel!=-1) draw_hilight(M,M->sel);
  XDestroyWindow(x_display, M->win);
  unregister_event_handler(M->win, MotionNotify, NULL);
  unregister_event_handler(M->win, ButtonPress, NULL);
  unregister_event_handler(M->win, ButtonRelease, NULL);
  unregister_event_handler(M->win, Expose, NULL);
  unregister_event_handler(M->parent, ButtonPress, M->parent_click_handler);
  unregister_event_handler(M->parent, ButtonRelease, M->parent_release_handler);

  if (M->func)
    (*M->func)(idx, M->arg);
}


static void hilight_item(menu_t *M, int x, int y)
{
  int k;
  if (x<0 || x>=M->w) return ;
  k= menu_y2i(M,y);
  if (k<0 || k>=M->n_items) return ;
  if (k==M->sel) return ;
  if (M->sel!=-1) draw_hilight(M,M->sel);
  M->sel= k;
  draw_hilight(M,M->sel);
}

static int  menu_i2y(menu_t* M, int idx)
{
  return M->itemh*idx + MO_pad_ver;
}
static int  menu_y2i(menu_t* M, int y)
{
  return (y-MO_pad_ver)/M->itemh;
}
static void draw_hilight(menu_t *M,int idx)
{
  int y;
  y= menu_i2y(M,idx);
  xor_rect(M->image->data,M->w,M->h,
           0, y, M->w, M->itemh, MO_hilight);
  XClearArea(x_display, M->win, 0,y, M->w, M->itemh, True);
}

  // this should be called only when the  menu is inactive. 
  // otherwise we would have Windows hanging around etc.
void menu_free(menu_t *M)
{
  if (M->image) destroy_image(M->image);
  free(M);
}

